// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.cards.
// 
// smithers.cards is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.cards is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.cards. If not, see <http://www.gnu.org/licenses/>.

package smithers.cards;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

/**
 * Plays Klondike in a text terminal.
 */
public class TerminalKlondike {
    
    private static BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));
    private static Klondike game;
    
    private TerminalKlondike() {}
    
    /**
     * Plays a game of Klondike. Reads moves from standard input and prints the
     * game's status on standard output. Valid inputs are:
     * <ul>
     * <li>{@code 'd'} - draw cards from the deck.</li>
     * <li>{@code 'q'} - quit the game.</li>
     * <li><em>nm</em> - move card(s) from stack <em>n</em> (0 for wastepile,
     * 1-7 for tableau) to stack <em>m</em> (0 for foundations, 1-7 for
     * tableau).</li>
     * </ul>
     * 
     * @param args the arguments, which are all ignored
     */
    public static void main(String[] args) throws IOException {
        if (args.length > 0) for (String arg : args) if (arg.equals("-h") || arg.equals("--help")) {
            System.out.println("smithers.cards.TerminalKlondike -- play Klondike in a terminal.");
            System.out.println("Usage:  java smithers.cards.TerminalKlondike [-h|--help]");
            System.out.println("If -h or --help is given, print this help and exit, otherwise play a game of solitare.");
            System.out.println("Commands:");
            System.out.println("  d     Draw cards (3 at a time) from the deck.");
            System.out.println("  nm    (where n and m are digits from 0 to 7)");
            System.out.println("        Move a card or cards from the postion n to position m. Positions 1-7 are the piles in the tableau from left to right. " +
                               "n=0 means move from the wastepile. m=0 means move to the foundations.");
            System.out.println("  q     Quit the game.");
            return;
        }
        game = new Klondike();
        show();
        theMainLoop();
    }
    
    private static void show() {
        String printThis;
        Card card;
        int i, j;
        
        System.out.println();
        System.out.println();
        
        // Show deck, waste and foundations:
        printThis = (game.deckLength() == 0) ? ((game.drawsLeft() == 0) ? "XX " : "OO ") : "## ";
        Card[] waste = game.visibleWaste();
        if (waste.length == 0) {
            printThis += "--    ";
        } else {
            printThis += waste[0] + "    ";
        }
        for (i=0; i<4; i++) {
            card = game.foundationTopCard(i);
            if (card == null) {
                printThis += "-- ";
            } else {
                printThis += card + " ";
            }
        }
        System.out.println(printThis);
        for (i=1; i<3; i++) {
            if (waste.length < i+1) {
                System.out.println();
            } else {
                System.out.println("   " + waste[i]);
            }
            
        }
        System.out.println();
        
        // Show tableau:
        Card[][] tableau = new Card[7][];
        int maxHeight = 0;
        for (i=0; i<7; i++) {
            tableau[i] = game.tableauCards(i);
            j = game.faceDownLength(i) + tableau[i].length;
            if (j > maxHeight) maxHeight = j;
        }
        
        for (j=0; j<maxHeight; j++) {
            printThis = "";
            for (i=0; i<7; i++) {
                if (game.faceDownLength(i) > j) {
                    printThis += "## ";
                } else if (game.faceDownLength(i) + tableau[i].length > j) {
                    printThis += tableau[i][j - game.faceDownLength(i)] + " ";
                } else {
                    printThis += "   ";
                }
            }
            System.out.println(printThis);
        }
        
        if (game.won()) {
            System.out.println("Congratulations, you win!");
        }
    }
    
    private static void theMainLoop() throws IOException {
        String kbdIn = "";
        int from, to, i;
        boolean show;
        
        mainloop:
        while (true) {
            kbdIn = stdin.readLine();
            show = true;
            for (i=0; i<kbdIn.length(); i++) {
                if (kbdIn.charAt(i) == ' ') { // Nothing
                    
                } else if (kbdIn.charAt(i) == 'd' || kbdIn.charAt(i) == '/') { // Draw
                    try {
                        game.draw();
                    } catch (CardMoveException e) {
                        System.out.println("Sorry, no draws left.");
                        show = false;
                    }
                    
                } else if (kbdIn.charAt(i) == '*') { // TODO - put all available cards up
                    System.out.println("TODO - put all available cards up.");
                    show = false;
                    
                } else if (kbdIn.charAt(i) == 'n') { // TODO - start new game
                    System.out.println("TODO - start new game.");
                    show = false;
                    
                } else if (kbdIn.charAt(i) == 'q') { // Exit game
                    System.out.println("Bye Bye.");
                    return;
                    
                } else if (kbdIn.charAt(i) >= '0' && kbdIn.charAt(i) <= '7') { // Move cards
                    try {
                        from = Integer.parseInt(kbdIn.substring(i, i+1));
                        to = Integer.parseInt(kbdIn.substring(i+1, i+2));
                        i++;
                        try {
                            game.move(from, to);
                        } catch (Exception e) {
                            System.out.println("Invalid move:");
                            System.out.println(e.getMessage());
                            show = false;
                        }
                    } catch (Exception e) {
                        System.out.println("Bad input from \"" + kbdIn.substring(i) + "\"");
                        continue mainloop;
                    }
                    
                } else {
                    System.out.println("Bad input from \"" + kbdIn.substring(i) + "\"");
                    continue mainloop;
                    
                }
            }
            if (show) show();
        }
    }
    
}
