// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.IOException;

/**
 * Implementer for the brainfuck programming language. This implementation
 * defaults to the following: <br />
 * The program is 30000 {@code char}s, the memory is 30000 {@code byte}s. The
 * program is read from a file, input and output are on {@code System.in} and
 * {@code System.out}. The data cells are 1 byte in size, and wrap around.
 * <br />
 * This version adds an extra command, '#', which dumps the memory to output in
 * hexadecimal. <br />
 * Some of the previous can be changed however. See the contructor
 * documentation for details.
 */
public class Brainfuck implements Runnable {
    
    private int progLen = 30000;
    private int dataLen = 30000;
    private byte[] data;
    private char[] prog;
    private int i, p, iMax, pLen;
    private int[] bracketA, bracketB;
    private boolean useDump = true;
    
    /**
     * Creates a new instance with the program from a file and some of the
     * options changed.
     * 
     * @param progLen the length of the program array
     * @param dataLen the length of the data array
     * @param useDump whether to use '#' to dump the data array
     * @param sourceFile the file to read the program from
     */
    public Brainfuck(int progLen, int dataLen, boolean useDump, String sourceFile) throws IOException {
        this.progLen = progLen;
        this.dataLen = dataLen;
        this.useDump = useDump;
        data = new byte[dataLen];
        prog = new char[progLen];
        pLen = new InputStreamReader(new FileInputStream(sourceFile)).read(prog, 0, progLen);
        removeComments();
        cacheBrackets();
        i = 0;
        p = 0;
    }
    
    /**
     * Creates a new instance with the program from a file.
     * 
     * @param sourceFile the file to read the program from
     */
    public Brainfuck(String sourceFile) throws IOException {
        data = new byte[dataLen];
        prog = new char[progLen];
        pLen = new InputStreamReader(new FileInputStream(sourceFile)).read(prog, 0, progLen);
        removeComments();
        cacheBrackets();
        i = 0;
        p = 0;
    }
    
    private void removeComments() {
        int i, d=0;
        for (i=0; i<pLen; i++) {
            if (prog[i]=='<' || prog[i]=='>' || prog[i]=='-' || prog[i]=='+' || prog[i]=='.' || prog[i]==',' || prog[i]=='[' || prog[i]==']' || (prog[i]=='#' && useDump))
                prog[i-d] = prog[i];
            else
                d++;
        }
        pLen -= d;
    }
    
    private void cacheBrackets() {
        int i, j=0, k, bCount=0;
        for (i=0; i<pLen; i++) if (prog[i]=='[' || prog[i]==']') bCount++;
        bracketA = new int[bCount];
        bracketB = new int[bCount];
        for (i=0; i<pLen; i++) {
            if (prog[i]=='[') {
                bracketA[j] = i;
                bracketB[j] = -1;
                j++;
            } else if (prog[i]==']') {
                k=j-1;
                while (bracketB[k] > -1) k--;
                bracketA[j] = bracketB[k] = i;
                bracketB[j] = bracketA[k];
                j++;
            }
        }
    }
    
    /**
     * Runs the program until it finishes.
     */
    public void run() {
        while (p<pLen) step();
        System.out.println();
    }
    
    /**
     * Executes a single step.
     */
    public void step() {
        if (p>=pLen) return;
        try {
            switch (prog[p]) {
                case '<': i--; break;
                case '>': i++; break;
                case '-': data[i]--; break;
                case '+': data[i]++; break;
                case '.': System.out.write(data[i]); break;
                case ',': data[i] = (byte)System.in.read(); break;
                case '[': if (data[i]==0) skip(); break;
                case ']': if (data[i]!=0) skip(); break;
                case '#': memDump(); break;
            }
            if (i>iMax) iMax = i;
            p++;
        } catch (Exception e) {
            System.out.println("\nError: " + e.getMessage());
            System.exit(1);
        }
    }
    
    private void memDump() {
        for (int j=0; j<=iMax; j++)
            System.out.println(Integer.toString((data[j]+256)%256, 16) + (i==j ? " <" : ""));
        System.out.println("---");
    }
    
    private void skip() {
        int a=0, b=bracketA.length, x=b/2;
        while (bracketA[x] != p) {
            if (bracketA[x] < p) {
                a=x+1;
            } else {
                b=x;
            }
            x=(a+b)/2;
        }
        p=bracketB[x];
    }
    
    /**
     * Main method, evaluates the arguments and runs the program in the
     * specified file. The arguments are as follows: <br />
     * <table>
     *   <hr>
     *     <th>Argument</th>
     *     <th>Meaning</th>
     *   </hr>
     *   <tr>
     *     <td>-p{@code N}</td>
     *     <td>Sets the length of the program array to {@code N}.</td>
     *   </tr>
     *   <tr>
     *     <td>-d{@code N}</td>
     *     <td>Sets the length of the data array to {@code N}.</td>
     *   </tr>
     *   <tr>
     *     <td>--nodump</td>
     *     <td>Turns off the use of '#' to dump the data array.</td>
     *   </tr>
     *   <tr>
     *     <td>(Anything else)</td>
     *     <td>Sets the file name.</td>
     *   </tr>
     * </table>
     * 
     * @param args the arguments
     */
    public static void main(String[] args) throws IOException {
        int progLen = 30000;
        int dataLen = 30000;
        boolean useDump = true;
        String fileName = null;
        for (int i=0; i<args.length; i++) {
            if (args[i].startsWith("-p"))
                progLen = Integer.parseInt(args[i].substring(2));
            else if (args[i].startsWith("-d"))
                dataLen = Integer.parseInt(args[i].substring(2));
            else if (args[i].equals("--nodump"))
                useDump = false;
            else if (args[i].equals("-h") || args[i].equals("--help")) {
                System.out.println("smithers.extras.Brainfuck -- brainfuck interpreter.");
                return;
            } else
                fileName = args[i];
        }
        new Brainfuck(progLen, dataLen, useDump, fileName).run();
    }
    
}
