// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.java3d.
// 
// smithers.java3d is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.java3d is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.java3d. If not, see <http://www.gnu.org/licenses/>.

package smithers.java3d;

import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import smithers.algorithms.Vector;

/**
 * Represents a recangular image arbitrarily suspended in 3-dimensional space.
 */
public class Image3D {
    
    private BufferedImage image;
    private float[] base;
    private float[] right, down, front;
    private Polygon3D clip;
    
    public Image3D(BufferedImage image, float[] topLeft, float[] topRight, float[] bottomLeft) {
        this(image, topLeft, topRight, bottomLeft, null);
    }
    
    public Image3D(BufferedImage image, float[] topLeft, float[] topRight, float[] bottomLeft, Polygon3D clip) {
        this.image = image;
        this.base = topLeft;
        this.right = Vector.difference(topRight, topLeft);
        this.down = Vector.difference(bottomLeft, topLeft);
        this.front = Vector.vectorProduct(down, right);
        this.clip = clip;
        if (front[0] == 0 && front[1] == 0 && front[2] == 0) throw new IllegalArgumentException("The image's 3D location is degenerate.");
    }
    
    public void paint(Graphics2D g, Transformation trans) {
        try {
            BufferedImage src = image;
            Rectangle srcBounds = new Rectangle(0, 0, src.getWidth(), src.getHeight());
            int[] srcPixels = src.getRGB(0, 0, srcBounds.width, srcBounds.height, null, 0, srcBounds.width);
            
            Rectangle dstBounds = g.getClipBounds();
            if (dstBounds != null)
                dstBounds = dstBounds.intersection(getBounds(trans));
            else
                dstBounds = getBounds(trans);
            BufferedImage dst = createDestImage(dstBounds);
            int[] dstPixels = dst.getRGB(0, 0, dstBounds.width, dstBounds.height, null, 0, dstBounds.width);
            Polygon dstClip = null;
            if (clip != null) {
                float[][] temp = trans.apply(clip.points(), null);
                int[] x = new int[temp.length];
                int[] y = new int[temp.length];
                for (int i=0; i<temp.length; i++) {
                    x[i] = (int)temp[i][0];
                    y[i] = (int)temp[i][1];
                }
                dstClip = new Polygon(x, y, temp.length);
            }
            
            Point2D dstPoint, srcPoint;
            Shape gClip = g.getClip();
            for (int i=0; i<dstBounds.width; i++) for (int j=0; j<dstBounds.height; j++) {
                dstPoint = new Point(dstBounds.x + i, dstBounds.y + j);
                if ((gClip == null || gClip.contains(dstPoint)) && (dstClip==null || dstClip.contains(dstPoint))) {
                    srcPoint = untransform(dstPoint, trans);
                    Point intPt = new Point((int)Math.floor(srcPoint.getX()), (int)Math.floor(srcPoint.getY()));
                    if (srcBounds.contains(intPt)) dstPixels[i + j * dstBounds.width] = srcPixels[intPt.x + intPt.y * srcBounds.width];
                }
            }
            
            dst.setRGB(0, 0, dstBounds.width, dstBounds.height, dstPixels, 0, dstBounds.width);
            g.drawImage(dst, dstBounds.x, dstBounds.y, null);
        } catch (IllegalArgumentException e) {} // We get an exception in Line3D.intersectPlane() if the image plane goes through the camera
    }
    
    private Rectangle getBounds(Transformation trans) {
        Rectangle2D bounds = getBounds2D(trans);
        return new Rectangle((int)Math.floor(bounds.getMinX()),
                             (int)Math.floor(bounds.getMinY()),
                             (int)Math.floor(bounds.getWidth())+1,
                             (int)Math.floor(bounds.getHeight())+1);
    }
    
    private Rectangle2D getBounds2D(Transformation trans) {
        float[][] pts = new float[4][];
        pts[0] = base;
        pts[1] = Vector.sum(base, right);
        pts[2] = Vector.sum(pts[1], down);
        pts[3] = Vector.sum(base, down);
        
        float xMin = Float.POSITIVE_INFINITY, xMax = Float.NEGATIVE_INFINITY, yMin = Float.POSITIVE_INFINITY, yMax = Float.NEGATIVE_INFINITY;
        for (float[] p : trans.apply(pts, null)) {
            if (p[0] < xMin) xMin = p[0];
            if (p[0] > xMax) xMax = p[0];
            if (p[1] < yMin) yMin = p[1];
            if (p[1] > yMax) yMax = p[1];
        }
        return new Rectangle2D.Float(xMin, yMin, xMax-xMin, yMax-yMin);
    }
    
    private BufferedImage createDestImage(Rectangle bounds) {
        ColorModel destCM = ColorModel.getRGBdefault();
        return new BufferedImage(destCM,
                                 destCM.createCompatibleWritableRaster(bounds.width, bounds.height),
                                 destCM.isAlphaPremultiplied(),
                                 null);
    }
    
    private Point2D untransform(Point2D dst, Transformation trans) {
        float[] temp = {(float)dst.getX(), (float)dst.getY()};
        float[][] line = trans.invert(temp);
        Line3D l = new Line3D(line[0], Vector.difference(line[1], line[0]));
        temp = l.intersectPlane(base, front);
        temp = Vector.difference(temp, base);
        float s = Vector.scalarProduct(temp, right) / Vector.scalarProduct(right, right);
        float t = Vector.scalarProduct(temp, down) / Vector.scalarProduct(down, down);
        return new Point2D.Float(s * image.getWidth(), t * image.getHeight());
    }
    
}
