// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.java3d.
// 
// smithers.java3d is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.java3d is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.java3d. If not, see <http://www.gnu.org/licenses/>.

package smithers.java3d;

import java.awt.AWTEvent;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

/**
 * Class which draws a 3D object on an AWT {@code Canvas}. Objects of this
 * class can have {@link Drawable3D} objects added to them, which will then be
 * displayed, and have mouse events dispatched to their registered listeners.
 * <br />
 * The semantics of mouse events are as follows:<br />
 * On a press, click or move event the
 * <span style="font-style: italic;">current object</span> is set to the object
 * currently under the cursor, or {@code null} if there is no such object. Then
 * the event is dispatched to the listeners registered with the
 * <span style="font-style: italic;">current object</span> if this is not
 * {@code null}. Wheel events are always dispatched to the object under the
 * cursor, without setting the <span style="font-style: italic;">current
 * object</span>. Enter and exit events are currently not dispatched to any of
 * the objects on the panel.<br />
 * If the event is not consumed, it is then dispatched to the listeners
 * registered with the panel. This includes a listener specified by the panel
 * itself, which rotates the view when the cursor is dragged.
 */
public abstract class Panel3D extends Canvas {
    
    private class MyListener implements MouseListener, MouseMotionListener {
        
        private int mouseX = 0, mouseY = 0;
        
        public void mousePressed(MouseEvent e) {
            mouseX = e.getX();
            mouseY = e.getY();
        }
        
        public void mouseDragged(MouseEvent e) {
            float x = (mouseX - e.getX()) / 50f;
            float y = (e.getY() - mouseY) / 50f;
            rotateView(y, x, 0);
            mouseX = e.getX();
            mouseY = e.getY();
            repaint();
        }
        
        /* Specified by MouseListener */
        public void mouseClicked(MouseEvent e) {}
        public void mouseReleased(MouseEvent e) {}
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        
        /* Specified by MouseMotionListener */
        public void mouseMoved(MouseEvent e) {}
        
    }
    
    /**
     * A simple drawable object which draws and fills a {@code Polygon3D}.
     */
    public static class DrawablePolygon extends Drawable3D {
        private Color draw, fill;
        
        /**
         * Creates a new {@code DrawablePolygon} with the specified polygon and
         * draw and fill colours. The colours can be {@code null}, which
         * indicates not to perform the corresponding operation.
         * 
         * @param poly the polygon to draw
         * @param draw the colour to outline the polygon in
         * @param fill the colour to fill the polygon with
         */
        public DrawablePolygon(Polygon3D poly, Color draw, Color fill) {
            this.poly = poly;
            this.draw = draw;
            this.fill = fill;
        }
        
        /**
         * {@inheritDoc}
         */
        public Shape drawOn(Graphics2D g, Transformation trans) {
            float[][] temp3d = poly.points();
            float[][] temp2d = trans.apply(temp3d, null);
            Polygon poly2d = new Polygon(new int[temp2d.length], new int[temp2d.length], 0);
            for (float[] p : temp2d) poly2d.addPoint((int)p[0], (int)p[1]);
            if (fill != null) {
                g.setColor(fill);
                g.fillPolygon(poly2d);
            }
            if (draw != null) {
                g.setColor(draw);
                g.drawPolygon(poly2d);
            }
            return poly2d;
        }
    }
    
    private List<Drawable3D> objects = new ArrayList<Drawable3D>();
    private Transformation perspective;
    private Transformation rotation;
    private Color background = Color.GRAY;
    
    private Shape[] objectShapes = null;
    private Drawable3D currentMouseEventObject = null;
    
    /**
     * Creates a new {@code Panel3D} object with the specified size.
     * 
     * @param size the object's size
     */
    public Panel3D(float size) {
        perspective = Transformation.perspective(3);
        perspective.preCompose(Transformation.translation(0, 0, size));
        rotation = Transformation.rotation3DFromIntrinsicXYZ((float)(Math.PI*7/6), (float)(Math.PI/3), 0);
        MyListener listener = new MyListener();
        addMouseListener(listener);
        addMouseMotionListener(listener);
        enableEvents(AWTEvent.MOUSE_EVENT_MASK | AWTEvent.MOUSE_MOTION_EVENT_MASK | AWTEvent.MOUSE_WHEEL_EVENT_MASK);
    }
    
    public void add(Drawable3D d) {
        objects.add(d);
    }
    
    public void remove(Drawable3D d) {
        objects.remove(d);
    }
    
    public void setBackground(Color c) {
        background = c;
    }
    
    public void rotateView(float a, float b, float c) {
        rotation.postCompose(Transformation.rotation3DFromIntrinsicXYZ(a, b, c));
    }
    
    public void update(Graphics g) {
        paint(g);
    }
    
    public void paint(Graphics g) {
        int w = g.getClipBounds().width;
        int h = g.getClipBounds().height;
        int size = h < w ? h : w;
        Transformation trans = Transformation.composition(perspective, rotation);
        trans.postCompose(Transformation.scale(2, size));
        trans.postCompose(Transformation.translation(w/2, h/2));
        
        int n = objects.size(), i;
        double[] objectDepths = new double[n];
        for (i=0; i<n; i++) objectDepths[i] = trans.getFactor(objects.get(i).poly.vertexCentroid());
        int nextI;
        for (i=n-1; i>0; i=nextI) {
            nextI = 0;
            for (int j=0; j<i; j++) {
                if (objectDepths[j+1] < objectDepths[j]) {
                    Drawable3D tempObj = objects.get(j);
                    objects.set(j, objects.get(j+1));
                    objects.set(j+1, tempObj);
                    double tempDepth = objectDepths[j];
                    objectDepths[j] = objectDepths[j+1];
                    objectDepths[j+1] = tempDepth;
                    nextI = j;
                }
            }
        }
        
        BufferedImage bufferImage = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g1 = bufferImage.createGraphics();
        g1.setColor(background);
        g1.fillRect(0, 0, w, h);
        
        objectShapes = new Shape[n];
        for (i=n-1; i>=0; i--)
            if (objectDepths[i] >= 0) objectShapes[i] = objects.get(i).drawOn(g1, trans);
        
        g.drawImage(bufferImage, 0, 0, null);
    }
    
    private Drawable3D getMouseEventObject(Point p) {
        for (int i=0; i<objectShapes.length; i++)
            if (objectShapes[i] != null && objectShapes[i].contains(p))
                return objects.get(i);
        return null;
    }
    
    private void setCurrentMouseEventObject(Point p) {
        currentMouseEventObject = getMouseEventObject(p);
    }
    
    protected void processMouseEvent(MouseEvent e) {
        int eventID = e.getID();
        if (eventID == MouseEvent.MOUSE_CLICKED || eventID == MouseEvent.MOUSE_PRESSED)
            setCurrentMouseEventObject(e.getPoint());
        if (currentMouseEventObject != null) {
            Object eventSource = e.getSource();
            e.setSource(currentMouseEventObject);
            switch (eventID) {
                case MouseEvent.MOUSE_CLICKED:
                    for (MouseListener l : currentMouseEventObject.mouseListeners) l.mouseClicked(e);
                    break;
                case MouseEvent.MOUSE_PRESSED:
                    for (MouseListener l : currentMouseEventObject.mouseListeners) l.mousePressed(e);
                    break;
                case MouseEvent.MOUSE_RELEASED:
                    for (MouseListener l : currentMouseEventObject.mouseListeners) l.mouseReleased(e);
                    break;
                case MouseEvent.MOUSE_ENTERED:
                    break;
                case MouseEvent.MOUSE_EXITED:
                    break;
            }
            e.setSource(eventSource);
        }
        if (!e.isConsumed()) {
            super.processMouseEvent(e);
        }
    }
    
    protected void processMouseMotionEvent(MouseEvent e) {
        int eventID = e.getID();
        if (eventID == MouseEvent.MOUSE_MOVED)
            setCurrentMouseEventObject(e.getPoint());
        if (currentMouseEventObject != null) {
            Object eventSource = e.getSource();
            e.setSource(currentMouseEventObject);
            switch (eventID) {
                case MouseEvent.MOUSE_DRAGGED:
                    for (MouseMotionListener l : currentMouseEventObject.mouseMotionListeners) l.mouseDragged(e);
                    break;
                case MouseEvent.MOUSE_MOVED:
                    for (MouseMotionListener l : currentMouseEventObject.mouseMotionListeners) l.mouseMoved(e);
                    break;
            }
            e.setSource(eventSource);
        }
        if (!e.isConsumed()) {
            super.processMouseMotionEvent(e);
        }
    }
    
    protected void processMouseWheelEvent(MouseWheelEvent e) {
        int eventID = e.getID();
        Drawable3D source = getMouseEventObject(e.getPoint());
        if (source != null) {
            Object eventSource = e.getSource();
            e.setSource(source);
            switch (eventID) {
                case MouseEvent.MOUSE_WHEEL:
                    for (MouseWheelListener l : currentMouseEventObject.mouseWheelListeners) l.mouseWheelMoved(e);
                    break;
            }
            e.setSource(eventSource);
        }
        if (!e.isConsumed()) {
            super.processMouseWheelEvent(e);
        }
    }
    
}
