// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

/**
 * Class to store the state of a hanjie puzzle.
 */
public class HanjieState implements PuzzleState<Hanjie> {
    
    Hanjie puzzle;
    int[][] state;
    
    /**
     * Constructs new solver for the specified puzzle.
     * 
     * @param puzzle the puzzle
     */
    public HanjieState(Hanjie puzzle) {
        this.puzzle = puzzle.isMutable() ? puzzle.makeImmutableCopy() : puzzle;
        state = new int[this.puzzle.height][this.puzzle.width];
        for (int i=0; i<this.puzzle.height; i++) for (int j=0; j<this.puzzle.width; j++)
            state[i][j] = -1;
    }
    
    public Hanjie getPuzzle() {
        return puzzle;
    }
    
    /**
     * Gets the state of a specificate pixel.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @return the state of the pixel in row {@code row}, column {@code column}
     */
    public int getState(int row, int column) {
        if (row < 0 || column < 0 || row >= puzzle.height || column >= puzzle.width)
            throw new IllegalArgumentException("Invalid location: " + row + ", " + column + ".");
        return state[row][column];
    }
    
    /**
     * Adds to the attempt at solving the puzzle.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @param state  the index of the colour to set coloured, 0 for not
     *               coloured, or -1 to reset
     */
    public void setState(int row, int column, int state) {
        if (row < 0 || column < 0 || row >= puzzle.height || column >= puzzle.width)
            throw new IllegalArgumentException("Invalid location: " + row + ", " + column + ".");
        if (state < -1 || (puzzle.hasColour ? state >= puzzle.getColourCount() : state > 1))
            throw new IllegalArgumentException("State is invalid.");
        this.state[row][column] = state;
    }
    
    /**
     * Clears the state of the puzzle.
     */
    public void clearState() {
        for (int i=0; i<puzzle.height; i++) for (int j=0; j<puzzle.width; j++)
            state[i][j] = -1;
    }
    
    // TODO: check that it matches the clues
    public boolean isSolved() {
        boolean hasColour = puzzle.hasColour();
        int[] clue, clueColour = null;
        int line, clueID, cellID, colVal, i, j;
        
        for (line=0; line<puzzle.height; line++)
        {
            clue = puzzle.getRowClue(line);
            if (hasColour) clueColour = puzzle.getRowColour(line);
            i = 0;
            for (clueID=0; clueID<clue.length; clueID++)
            {
                colVal = hasColour ? clueColour[clueID] : 1;
                while (i < puzzle.width && state[line][i] <= 0)
                    i++;
                j = i;
                while (i < puzzle.width && state[line][i] == colVal)
                    i++;
                if (i - j != clue[clueID])
                    return false;
            }
            while (i < puzzle.width)
            {
                if (state[line][i] > 0)
                    return false;
                i++;
            }
        }
        for (line=0; line<puzzle.width; line++)
        {
            clue = puzzle.getColumnClue(line);
            if (hasColour) clueColour = puzzle.getColumnColour(line);
            i = 0;
            for (clueID=0; clueID<clue.length; clueID++)
            {
                colVal = hasColour ? clueColour[clueID] : 1;
                while (i < puzzle.height && state[i][line] <= 0)
                    i++;
                j = i;
                while (i < puzzle.height && state[i][line] == colVal)
                    i++;
                if (i - j != clue[clueID])
                    return false;
            }
            while (i < puzzle.height)
            {
                if (state[i][line] > 0)
                    return false;
                i++;
            }
        }
        return true;
    }
    
    /**
     * Generates a {@code String} representing the current situation of the
     * puzzle. Represents unknown as a grave accent '`', white (unfilled) as a
     * space ' ' and colour 1 as a hash '#'. In colour puzzles, for {@code i>1}
     * it represents colour {@code i} as {@code (char)(0x0022+i)}, i.e. Unicode
     * characters continuing from '#'.
     * 
     * @return a {@code String} representing the state
     */
    public String toString() {
        char[] result = new char[(puzzle.width + 1) * puzzle.height];
        for (int i=0; i<puzzle.height; i++) {
            for (int j=0; j<puzzle.width; j++) {
                result[i * (puzzle.width + 1) + j] = state[i][j] < 0 ? '`' : state[i][j] == 0 ? ' ' : (char)(0x22 + state[i][j]);
            }
            result[i * (puzzle.width + 1) + puzzle.width] = '\n';
        }
        return new String(result);
    }
    
}
