// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;

/**
 * Makes mosaic puzzles from images.
 */
public class MosaicMaker {
    
    static int height;
    static int width;
    static int[][] grid;
    
    private MosaicMaker() {}
    
    private static int getClue(int i0, int j0) {
        int result = 0;
        for (int i = i0>0 ? i0-1 : 0; i<=i0+1 && i<height; i++)
            for (int j = j0>0 ? j0-1 : 0; j<=j0+1 && j<width; j++)
                if (grid[i][j] == 1) result++;
        return result;
    }
    
    public static void main(String[] args) throws IOException {
        if (args.length == 0 || args[0].equals("-h") || args[0].equals("--help")) {
            System.out.println("smithers.logicpuzzles.MosaicMaker -- Creates mosaic puzzles from images.");
            System.out.println("Usage:");
            System.out.println("        java smithers.logicpuzzles.MosaicMaker -h|--help");
            System.out.println("        java smithers.logicpuzzles.MosaicMaker image");
            System.out.println("");
            System.out.println("If -h or --help is given, display this help and exit.");
            System.out.println("Otherwise image must be the path to an image from which to create a mosaic puzzle. The puzzle will be printed to standard output.");
            System.out.println("The image should be of the same size as the puzzle to be produced, and all non-white pixels are treated as black.");
            return;
        }
        
        int i,j,k;
        
        //Read the input file
        boolean advanced = args[0].equals("-a");
        BufferedImage image = ImageIO.read(new File(args[advanced?1:0]));
        height = image.getHeight();
        width = image.getWidth();
        int[] temp = image.getRGB(0, 0, width, height, null, 0, width);
        
        grid = new int[height][width];
        for (i=0; i<height; i++) for (j=0; j<width; j++)
            grid[i][j] = temp[j+i*height]==0xffffffff ? 0 : 1;
        Mosaic puzzle = new Mosaic(height, width);
        for (i=0; i<height; i++) for (j=0; j<width; j++)
            puzzle.setClue(i, j, getClue(i, j));
        MosaicState state = new MosaicState(puzzle);
        MosaicSolver solver = new MosaicSolver(state, advanced);
        
        //Solve, catching used clues
        int[] clueI = new int[width*height+1];
        int[] clueJ = new int[width*height+1];
        k = 0;
        while (solver.step()) {
            temp = solver.getLastChange();
            for (i=0; i<temp.length/2; i++) {
                for (j=0; j<k; j++)
                    if (clueI[j] == temp[2*i] && clueJ[j] == temp[2*i+1]) break;
                if (j<k) continue;
                clueI[k] = temp[2*i];
                clueJ[k] = temp[2*i+1];
                k++;
            }
        }
        clueI[k] = -1;
        for (i=0; i<height; i++) for (j=0; j<width; j++) {
            if (state.getState(i, j) == -1) {
                System.err.println("No solvable puzzle could be made.");
                return;
            }
        }
        
        //Remake puzzle with only the clues used
        for (i=0; i<height; i++) for (j=0; j<width; j++)
            puzzle.setClue(i, j, -1);
        for (k=0; clueI[k]!=-1; k++)
            puzzle.setClue(clueI[k], clueJ[k], getClue(clueI[k], clueJ[k]));
        
        //Test without each clue
        for (k=0; clueI[k]!=-1; k++) {
            puzzle.setClue(clueI[k], clueJ[k], -1);
            if (!new MosaicSolver(puzzle, advanced).solve())
                puzzle.setClue(clueI[k], clueJ[k], getClue(clueI[k], clueJ[k]));
        }
        
        //Done - output result
        System.out.print(puzzle.toString());
    }
    
}
