// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.twisty.
// 
// smithers.twisty is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.twisty is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.twisty. If not, see <http://www.gnu.org/licenses/>.

package smithers.twisty;

import java.util.Random;
import smithers.java3d.Polygon3D;

/*
 U              0
LFRB           1234
 D              5
      +-----+
      | 0  1|
      | 2  3|
+-----+-----+-----+-----+
| 4  5| 8  9|12 13|16 17|
| 6  7|10 11|14 15|18 19|
+-----+-----+-----+-----+
      |20 21|
      |22 23|
      +-----+
*/

/**
 * Represents a <a href="http://en.wikipedia.org/wiki/Rubik's_Cube">Rubik's
 * Cube</a>, or similar larger/smaller twisty cube puzzle.
 */
public class RubiksCube implements StickerSwappingTwistyPuzzle, AnimatedTwistyPuzzle {
    
    private static class RubiksTwist extends Twist {
        
        int face, d1, d2, dir;
        
        RubiksTwist(int face, int d1, int d2, int dir) {
            this.face = face;
            this.d1 = d1;
            this.d2 = d2;
            this.dir = (dir % 4 + 6) % 4 - 2;
        }
        
    }
    
    private int s, ss; // ss == s * s
    private int[] sticker;
    private int[] location;
    private int[] rotation;
    
    private boolean animating = false;
    private double aniTime=0, aniCurrent=0;
    private RubiksTwist aniTwist = null;
    
    /**
     * Constructs a new 3x3x3 Rubik's Cube.
     */
    public RubiksCube() {
        this(3);
    }
    
    /**
     * Constructs a new cube with a given side length.
     * 
     * @param side the length of the side
     */
    public RubiksCube(int side) {
        s = side;
        ss = s*s;
        sticker = new int[ss*6];
        location = new int[ss*6];
        rotation = new int[ss*6];
        for (int i=0; i<ss*6; i++) {
            sticker[i] = i;
            location[i] = i;
            rotation[i] = 0;
        }
    }
    
/* Simple accessors */
    
    /**
     * Gets the side length.
     * 
     * @return the length of a side
     */
    public int side() {
        return s;
    }
    
    /**
     * {@inheritDoc}
     */
    public int stickerCount() {
        return ss*6;
    }
    
    /**
     * {@inheritDoc}
     */
    public int getColor(int index) {
        switch (index / ss) {
            case 0: return 0xffffffff; // White
            case 1: return 0xff00ff00; // Green
            case 2: return 0xffff0000; // Red
            case 3: return 0xff0000ff; // Blue
            case 4: return 0xffff9900; // Orange
            case 5: return 0xffffff00; // Yellow
        }
        return 0; // No such face
    }
    
    /**
     * {@inheritDoc}
     */
    public double getScale() {
        return 2*s;
    }
    
/* getPolygon(int) */
    
    /**
     * {@inheritDoc}
     */
    public Polygon3D getPolygon(int index) {
        float[][] result = new float[4][];
        result[0] = posOnSticker(index, 0, 0);
        result[1] = posOnSticker(index, 1, 0);
        result[2] = posOnSticker(index, 1, 1);
        result[3] = posOnSticker(index, 0, 1);
        if (isMoving(index)) {
            float angle = (float)(aniCurrent / aniTime * aniTwist.dir * Math.PI / 2);
            switch (aniTwist.face) {
                case 1:
                case 4:
                case 5:
                    angle = -angle;
                    break;
            }
            float cosa = (float)Math.cos(angle);
            float sina = (float)Math.sin(angle);
            switch (aniTwist.face) {
                case 1:
                case 3:
                    for (int i=0; i<4; i++) {
                        float temp = result[i][1] * cosa - result[i][2] * sina;
                        result[i][2] = result[i][1] * sina + result[i][2] * cosa;
                        result[i][1] = temp;
                    }
                    break;
                case 0:
                case 5:
                    for (int i=0; i<4; i++) {
                        float temp = result[i][2] * cosa - result[i][0] * sina;
                        result[i][0] = result[i][2] * sina + result[i][0] * cosa;
                        result[i][2] = temp;
                    }
                    break;
                case 2:
                case 4:
                    for (int i=0; i<4; i++) {
                        float temp = result[i][0] * cosa - result[i][1] * sina;
                        result[i][1] = result[i][0] * sina + result[i][1] * cosa;
                        result[i][0] = temp;
                    }
                    break;
            }
        }
        return new Polygon3D(result);
    }
    
    private float[] posOnSticker(int index, float x, float y) {
        int loc = location[index];
        float x1 = loc%s;
        float y1 = loc/s%s;
        float hs = s/2f;
        switch ((rotation[index]%4+4)%4) {
            case 0:
                x1 += x;
                y1 += y;
                break;
            case 1:
                x1 += y;
                y1 += 1-x;
                break;
            case 2:
                x1 += 1-x;
                y1 += 1-y;
                break;
            case 3:
                x1 += 1-y;
                y1 += x;
                break;
        }
        x1 = x1-hs;
        y1 = y1-hs;
        float[] result = new float[3];
        switch (loc/ss) {
            case 0: result[0] = x1; result[1] = hs; result[2] = y1; break;
            case 1: result[0] = -hs; result[1] = -y1; result[2] = x1; break;
            case 2: result[0] = x1; result[1] = -y1; result[2] = hs; break;
            case 3: result[0] = hs; result[1] = -y1; result[2] = -x1; break;
            case 4: result[0] = -x1; result[1] = -y1; result[2] = -hs; break;
            case 5: result[0] = x1; result[1] = -hs; result[2] = -y1; break;
        }
        return result;
    }
    
    //Works out whether a sticker is involved in the current twist.
    private boolean isMoving(int index) {
        if (!animating) return false;
        int depth = getDepth(aniTwist.face, location[index]);
        return aniTwist.d1 <= depth && depth < aniTwist.d2;
    }
    
    private int getDepth(int face, int loc) {
        int i = loc%s;
        int j = loc/s%s;
        int f = loc/ss;
        if (f == face) return 0;
        if (f == opposite(face)) return s-1;
        int x, y, z;
        switch (f) {
            case 0: x=i; y=s-1; z=j; break;
            case 1: x=0; y=s-1-j; z=i; break;
            case 2: x=i; y=s-1-j; z=s-1; break;
            case 3: x=s-1; y=s-1-j; z=s-1-i; break;
            case 4: x=s-1-i; y=s-1-j; z=0; break;
            case 5: x=i; y=0; z=s-1-j; break;
            default: assert false; return 0;
        }
        switch (face) {
            case 0: return s-1-y;
            case 1: return x;
            case 2: return s-1-z;
            case 3: return s-1-x;
            case 4: return z;
            case 5: return y;
            default: assert false; return 0;
        }
    }
    
/* Position getters */
    
    /**
     * Gets the location of a sticker. If the puzzle is partway through a twist,
     * this returns the value before the twist.
     * 
     * @param index the sticker index
     * @return the location index
     * @see #getSticker(int)
     * @see #getRotation(int)
     */
    public int getLocation(int index) {
        return location[index];
    }
    
    /**
     * Gets the index of the sticker in a specific location. If the puzzle is
     * partway through a twist, this returns the value before the twist.
     * 
     * @param loc the location index
     * @return the sticker index
     * @see #getLocation(int)
     * @see #getStickerRotation(int)
     */
    public int getSticker(int loc) {
        return sticker[loc];
    }
    
    /**
     * Gets the rotation of a sticker. The value returned is the rotation of
     * sticker {@code index} from 0 to 3. If the puzzle is partway through a
     * twist, this returns the value before the twist.
     * 
     * @param index the sticker index
     * @return the rotation of the sticker
     * @see #getLocation(int)
     */
    public int getRotation(int index) {
        return (rotation[index]%4+4)%4;
    }
    
    /**
     * Gets the rotation of the sticker in a location. The value returned is
     * the rotation of the sticker in in a location {@code loc} from 0 to 3. If
     * the puzzle is partway through a twist, this returns the value before the
     * twist.
     * 
     * @param loc the location of the sticker
     * @return the rotation of the sticker
     * @see #getSticker(int)
     */
    public int getStickerRotation(int loc) {
        return (rotation[sticker[loc]]%4+4)%4;
    }
    
/* Twists */
    
    /**
     * Gets the twist instance defined by the parameters. See {@link
     * #twist(int,int,int,int)} for the description of the twist.
     * 
     * @param face the face to twist parallel to
     * @param d1   the first layer to turn
     * @param d2   the first layer not to turn
     * @param dir  the direction
     * @throws IndexOutOfBoundsException if {@code d1<0 || d2<d1 || s<d2}
     */
    public Twist getTwist(int face, int d1, int d2, int dir) {
        if (d1<0) throw new IndexOutOfBoundsException("d1 must not be negative");
        if (d2<d1) throw new IndexOutOfBoundsException("d2 must not be less than d1");
        if (s<d2) throw new IndexOutOfBoundsException("d2 must not be more than " + s);
        return new RubiksTwist(face, d1, d2, dir);
    }
    
    /**
     * {@inheritDoc}
     */
    public Twist getTwist(int index, int dir) {
        if (index < 0) throw new IndexOutOfBoundsException("Index must be >=0");
        if (index >= ss*6) throw new IndexOutOfBoundsException("Index must be <" + ss*6);
        int loc = location[index];
        int x = loc%s;
        int y = loc/s%s;
        int f = loc/ss;
        dir = dir==0 ? 1 : dir==1 ? -1 : 2;
        
        if (y==0 && x>0 && x<s-1) return new RubiksTwist(side(f, 0, 0, 0)/ss, x, x+1, dir);
        else if (x==0 && y>0 && y<s-1) return new RubiksTwist(side(f, 1, 0, 0)/ss, s-y-1, s-y, dir);
        else if (y==s-1 && x>0 && x<s-1) return new RubiksTwist(side(f, 2, 0, 0)/ss, s-x-1, s-x, dir);
        else if (x==s-1 && y>0 && y<s-1) return new RubiksTwist(side(f, 3, 0, 0)/ss, y, y+1, dir);
        else return new RubiksTwist(f, 0, 1, dir);
    }
    
    /**
     * {@inheritDoc}
     */
    public void twist(Twist twist) {
        if (animating) throw new AnimatingException();
        RubiksTwist rt = (RubiksTwist)twist;
        twist(rt.face, rt.d1, rt.d2, rt.dir);
    }
    
    /**
     * {@inheritDoc}
     */
    public void twist(Twist twist, double time) {
        if (animating) throw new AnimatingException();
        RubiksTwist rt = (RubiksTwist)twist;
        if (time < 0.0) {
            time = Math.abs(time * rt.dir / 2);
        }
        if (time == 0.0) {
            twist(twist);
            return;
        }
        aniTwist = rt;
        animating = true;
        aniTime = time;
        aniCurrent = 0;
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean isAnimating() {
        return animating;
    }
    
    /**
     * {@inheritDoc}
     */
    public void advanceAnimation(double time) {
        if (animating) {
            aniCurrent += time;
            if (aniCurrent >= aniTime) {
                animating = false;
                twist(aniTwist);
            } else if (aniCurrent <= 0) {
                animating = false;
            }
        }
    }
    
    /**
     * Twists face {@code face} of the cube from layer {@code d1} (inclusive)
     * to layer {@code d2} (exclusive). Notice of course that the face itself
     * is turned iff {@code d1==0 && d2>0}, the opposite face is turned iff
     * {@code d2==s && d1<s}. Nothing happens if {@code d1==d2}. The direction
     * of the turn is specified by {@code dir} (mod 4): 1 = anticlockwise, 2 =
     * half twist, 3 = clockwise.
     * 
     * @param face the face to twist parallel to
     * @param d1   the first layer to turn
     * @param d2   the first layer not to turn
     * @param dir  the direction
     * @throws IndexOutOfBoundsException if {@code d1<0 || d2<d1 || s<d2}
     * @throws AnimatingException        if the puzzle is part way through an
     *                                   animation
     */
    public final void twist(int face, int d1, int d2, int dir) {
        if (animating) throw new AnimatingException();
        if (d1<0) throw new IndexOutOfBoundsException("d1 must not be negative");
        if (d2<d1) throw new IndexOutOfBoundsException("d2 must not be less than d1");
        if (s<d2) throw new IndexOutOfBoundsException("d2 must not be more than "+s);
        int i, j;
        //Move the face itself
        if (d1==0 && d2>0) {
            for (i=0; 2*i<s-1; i++) for (j=i; j<s-i-1; j++)
                rot(ss*face + s*i+j,
                    ss*face + s*j-i + s-1,
                    ss*face - s*i-j + ss-1,
                    ss*face - s*j+i + ss-s,
                    dir);
            if (s%2==1) rotation[sticker[ss*face + s*(s/2) + s/2]] += dir;
        }
        //Move the stickers on the sides
        for (i=0; i<s; i++) for (j=d1; j<d2; j++)
            rot(side(face, 3, i, j),
                side(face, 2, i, j),
                side(face, 1, i, j),
                side(face, 0, i, j),
                dir);
        //Move the opposite face
        if (d2==s && d1<s) {
            int opFace = opposite(face);
            for (i=0; i<s-1; i++) for (j=i; j<s-i-1; j++)
                rot(ss*opFace + s*i+j,
                    ss*opFace + s*j-i + s-1,
                    ss*opFace - s*i-j + ss-1,
                    ss*opFace - s*j+i + ss-s,
                    -dir);
            if (s%2==1) rotation[sticker[ss*opFace + s*(s/2) + s/2]] -= dir;
        }
    }
    
    /*
     * Rotates the stickers in locations la, lb, lc and ld.
     */
    private void rot(int la, int lb, int lc, int ld, int dir) {
        int fa = la/ss, fb = lb/ss, fc = lc/ss, fd = ld/ss;
        int sa = sticker[la], sb = sticker[lb], sc = sticker[lc], sd = sticker[ld];
        switch ((dir%4+4)%4) {
            case 1:
                location[sa] = ld;
                location[sb] = la;
                location[sc] = lb;
                location[sd] = lc;
                sticker[la] = sb;
                sticker[lb] = sc;
                sticker[lc] = sd;
                sticker[ld] = sa;
                rotation[sa] += getRelRot(fa, fd);
                rotation[sb] += getRelRot(fb, fa);
                rotation[sc] += getRelRot(fc, fb);
                rotation[sd] += getRelRot(fd, fc);
                break;
            case 2:
                location[sa] = lc;
                location[sb] = ld;
                location[sc] = la;
                location[sd] = lb;
                sticker[la] = sc;
                sticker[lb] = sd;
                sticker[lc] = sa;
                sticker[ld] = sb;
                rotation[sa] += getRelRot(fa, fd) + getRelRot(fd, fc);
                rotation[sc] += getRelRot(fc, fb) + getRelRot(fb, fa);
                rotation[sb] += getRelRot(fb, fa) + getRelRot(fa, fd);
                rotation[sd] += getRelRot(fd, fc) + getRelRot(fc, fb);
                break;
            case 3:
                location[sa] = lb;
                location[sb] = lc;
                location[sc] = ld;
                location[sd] = la;
                sticker[ld] = sc;
                sticker[lc] = sb;
                sticker[lb] = sa;
                sticker[la] = sd;
                rotation[sa] -= getRelRot(fb, fa);
                rotation[sb] -= getRelRot(fc, fb);
                rotation[sc] -= getRelRot(fd, fc);
                rotation[sd] -= getRelRot(fa, fd);
                break;
        }
    }
    
    /*
     * Works out how much a sticker's rotation should change if it is moved from
     * face1 to face2.
     */
    private int getRelRot(int face1, int face2) {
        if (face1 == face2) return 1;
        switch (face1) {
            case 0:
                switch (face2) {
                    case 1: return 1;
                    case 2: return 0;
                    case 3: return 3;
                    case 4: return 2;
                }
            case 1:
                switch (face2) {
                    case 0: return 3;
                    case 2: return 0;
                    case 4: return 0;
                    case 5: return 1;
                }
            case 2:
                switch (face2) {
                    case 0: return 0;
                    case 1: return 0;
                    case 3: return 0;
                    case 5: return 0;
                }
            case 3:
                switch (face2) {
                    case 0: return 1;
                    case 2: return 0;
                    case 4: return 0;
                    case 5: return 3;
                }
            case 4:
                switch (face2) {
                    case 0: return 2;
                    case 1: return 0;
                    case 3: return 0;
                    case 5: return 2;
                }
            case 5:
                switch (face2) {
                    case 1: return 3;
                    case 2: return 0;
                    case 3: return 1;
                    case 4: return 2;
                }
        }
        assert false;
        return 0;
    }
    
    /**
     * Looks at one of the faces next to face {@code top} as if it is forwards
     * and face {@code top} is on top, and gives the index of the sticker in
     * position {@code (x,y)}. {@code x} and {@code y} are interpreted modulo
     * the side length. The side is counted around anticlockwise by {@code
     * side} from the one on the left edge of face {@code top}.
     * 
     * @param top  the face to consider as the 'top'
     * @param side which side of the top face to look next to
     * @param x    the position on the face counting parallel to 'top'
     * @param y    the position on the face counting away from 'top'
     * @return the index of the required sticker
     */
    public final int side(int top, int side, int x, int y) {
        x %= s;
        y %= s;
        if (x<0) x += s;
        if (y<0) y += s;
        int face=-1;
        int dir=-1;
        switch (top) {
            case 0:
                switch (side) {
                    case 0: face=1; dir=0; break;
                    case 1: face=2; dir=0; break;
                    case 2: face=3; dir=0; break;
                    case 3: face=4; dir=0; break;
                }
                break;
            case 1:
                switch (side) {
                    case 0: face=4; dir=1; break;
                    case 1: face=5; dir=3; break;
                    case 2: face=2; dir=3; break;
                    case 3: face=0; dir=3; break;
                }
                break;
            case 2:
                switch (side) {
                    case 0: face=1; dir=1; break;
                    case 1: face=5; dir=0; break;
                    case 2: face=3; dir=3; break;
                    case 3: face=0; dir=2; break;
                }
                break;
            case 3:
                switch (side) {
                    case 0: face=2; dir=1; break;
                    case 1: face=5; dir=1; break;
                    case 2: face=4; dir=3; break;
                    case 3: face=0; dir=1; break;
                }
                break;
            case 4:
                switch (side) {
                    case 0: face=3; dir=1; break;
                    case 1: face=5; dir=2; break;
                    case 2: face=1; dir=3; break;
                    case 3: face=0; dir=0; break;
                }
                break;
            case 5:
                switch (side) {
                    case 0: face=1; dir=2; break;
                    case 1: face=4; dir=2; break;
                    case 2: face=3; dir=2; break;
                    case 3: face=2; dir=2; break;
                }
                break;
        }
        if (dir==0) return ss*face + s*y + x;
        if (dir==1) return ss*face + s*x - y + s - 1;
        if (dir==2) return ss*face - s*y - x + ss - 1;
        if (dir==3) return ss*face - s*x + y + ss - s;
        return -1;
    }
    
    /*
     * Gets the face opposite the specified face.
     */
    private int opposite(int face) {
        switch (face) {
            case 0: return 5;
            case 1: return 3;
            case 2: return 4;
            case 3: return 1;
            case 4: return 2;
            case 5: return 0;
        }
        throw new IndexOutOfBoundsException("face must be 0-5");
    }
    
/* Other stuff */
    
    /**
     * Scrambles the cube.
     */
    public void scramble() {
        // Scramble by applying random twists
        Random r = new Random();
        int parity = r.nextInt(2);
        for (int i=0; i<s*8+parity; i++) {
            int face = r.nextInt(6);
            int depth = r.nextInt(s);
            twist(face, depth, depth+1, 1);
        }
    }
    
    /**
     * Resets the cube.
     */
    public void reset() {
        for (int i=0; i<s*s*6; i++) {
            sticker[i] = i;
            location[i] = i;
            rotation[i] = 0;
        }
    }
    
    /**
     * Returns a {@code String} representation of this cube.
     * 
     * @return a {@code String} representing this cube
     */
    public String toString() {
        String result = "";
        String dashes = "", spaces = "";
        int i, j, k;
        for (i=0; i<s; i++) {
            dashes += "-";
            spaces += " ";
        }
        
        result += " " + spaces + "+" + dashes + "+\n";
        for (j=0; j<s; j++) {
            result += " " + spaces + "|";
            for (i=0; i<s; i++)
                result += sticker[i+s*j] / ss;
            result += "|\n";
        }
        result += "+" + dashes + "+" + dashes + "+" + dashes + "+" + dashes + "+\n";
        for (j=0; j<s; j++) {
            for (k=1; k<5; k++) {
                result += "|";
                for (i=0; i<s; ++i)
                    result += sticker[i+s*j+ss*k] / ss;
            }
            result += "|\n";
        }
        result += "+" + dashes + "+" + dashes + "+" + dashes + "+" + dashes + "+\n";
        for (j=0; j<s; j++) {
            result += " " + spaces + "|";
            for (i=0; i<s; i++)
                result += sticker[i+s*j+ss*5] / ss;
            result += "|\n";
        }
        result += " " + spaces + "+" + dashes + "+\n";
        return result;
    }
    
    /**
     * {@inheritDoc}
     * <br>
     * This puzzle is considered solved if each face of the cube contains
     * stickers of only one colour.
     * 
     * @return {@inheritDoc}
     */
    public boolean isSolved() {
        int temp;
        for (int f=0; f<6; f++) {
            temp = location[ss*f] / ss;
            for (int i=1; i<ss; i++) {
                if (location[ss*f+i] / ss != temp) return false;
            }
        }
        return true;
    }
    
    /**
     * Determines whether the puzzle is solved, including checking the
     * orientation of each sticker.
     * 
     * @return {@code true} if the puzzle is solved with all stickers correctly
     *                      oriented
     */
    public boolean isSolvedAligned() {
        int face, orient;
        for (int f=0; f<6; f++) {
            face = location[ss*f] / ss;
            orient = getRotation(ss*f);
            for (int i=1; i<s*s; i++) {
                if (location[ss*f+i] / ss != face) return false;
                if (getRotation(ss*f+i) != orient) return false;
            }
        }
        return true;
    }
    
}
