// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.twisty.
// 
// smithers.twisty is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.twisty is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.twisty. If not, see <http://www.gnu.org/licenses/>.

package smithers.twisty;

import java.util.Random;
import smithers.algorithms.Vector;
import smithers.java3d.Polygon3D;

/**
 * Represents a Rubik's Snake.
 */
public class Snake implements TwistyPuzzle {
    
    private static class SnakeTwist extends Twist {
        int piece, dir;
        
        SnakeTwist(int piece, int dir) {
            this.piece = piece;
            this.dir = (dir%4+6)%4-2;
        }
    }
    
    private final int n; //Number of pieces
    private final int[] position;
    
    /**
     * Creates a Rubik's Snake with 24 pieces.
     */
    public Snake() {
        this(24);
    }
    
    /**
     * Creates a Rubik's Snake-like puzzle with the specified number of pieces.
     * 
     * @param pieces the number of pieces
     */
    public Snake(int pieces) {
        n = pieces;
        position = new int[n-1];
        for (int i=0; i<n-1; i++) position[i] = 0;
    }
    
    /**
     * {@inheritDoc}
     */
    public int stickerCount() {
        return 5*n;
    }
    
    /**
     * {@inheritDoc}
     */
    public int getColor(int index) {
        return index/5%2==0 ? 0xff0000ff : 0xff00ff00;
    }
    
    /**
     * {@inheritDoc}
     */
    public double getScale() {
        return n;
    }
    
    /**
     * {@inheritDoc}
     */
    public Polygon3D getPolygon(int index) {
        if (index < 0) throw new IndexOutOfBoundsException("Index should be >=0");
        if (index >= 5*n) throw new IndexOutOfBoundsException("Index should be <" + 5*n);
        int piece = index/5;
        
        float[] center = {0, 0, 0};
        float[] in = {0, 1, 0};
        float[] out = {1, 0, 0};
        float[] temp = null;
        for (int i=0; i<piece; i++) {
            center[0] += out[0];
            center[1] += out[1];
            center[2] += out[2];
            switch (position[i]) {
                case 0: temp = in; break;
                case 1: temp = Vector.vectorProduct(in, out); break;
                case 2: temp = new float[] {-in[0], -in[1], -in[2]}; break;
                case 3: temp = Vector.vectorProduct(out, in); break;
            }
            in = out;
            out = temp;
        }
        float[] side = Vector.vectorProduct(in, out);
        
        float[] a = {out[0]/2, out[1]/2, out[2]/2};
        float[] b = {in[0]/2, in[1]/2, in[2]/2};
        float[] c = {side[0]/2, side[1]/2, side[2]/2};
        float[] d = {-out[0]/2, -out[1]/2, -out[2]/2};
        float[] e = {-in[0]/2, -in[1]/2, -in[2]/2};
        float[] f = {-side[0]/2, -side[1]/2, -side[2]/2};
        float[][] verts = new float[6][];
        for (int i=0; i<6; i++)
            verts[i] = Vector.sum(center, (i&1)==0 ? c : f, (i&2)==0 ? a : d, (i&4)==0 ? e : b);
        
        Polygon3D result;
        switch (index%5) {
            case 0: // in
                result = new Polygon3D(4);
                result.add(verts[0]);
                result.add(verts[2]);
                result.add(verts[3]);
                result.add(verts[1]);
                return result;
                
            case 1: // side1
                result = new Polygon3D(3);
                result.add(verts[0]);
                result.add(verts[4]);
                result.add(verts[2]);
                return result;
                
            case 2: // front
                result = new Polygon3D(4);
                result.add(verts[4]);
                result.add(verts[5]);
                result.add(verts[3]);
                result.add(verts[2]);
                return result;
                
            case 3: // side2
                result = new Polygon3D(3);
                result.add(verts[1]);
                result.add(verts[3]);
                result.add(verts[5]);
                return result;
                
            default: // out
                result = new Polygon3D(4);
                result.add(verts[0]);
                result.add(verts[1]);
                result.add(verts[5]);
                result.add(verts[4]);
                return result;
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public Twist getTwist(int index, int dir) {
        if (index < 0) throw new IndexOutOfBoundsException("Index should be >=0");
        if (index >= 5*n) throw new IndexOutOfBoundsException("Index should be <" + 5*n);
        if (index<5) return new SnakeTwist(0, 0);
        dir = dir==0 ? 1 : dir==1 ? -1 : 2;
        return new SnakeTwist(index/5, dir);
    }
    
    /**
     * {@inheritDoc}
     */
    public void twist(Twist twist) {
        SnakeTwist st = (SnakeTwist)twist;
        if (st.dir!=0) position[st.piece-1] = (position[st.piece-1] + st.dir + 4) % 4;
    }
    
    /**
     * {@inheritDoc}
     */
    public void scramble() {
        Random r = new Random();
        for (int i=0; i<n-1; i++) position[i] = r.nextInt(4);
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean isSolved() {
        for (int i=0; i<n-1; i++) if (position[i]!=0) return false;
        return true;
    }
    
}
