// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.twisty.
// 
// smithers.twisty is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.twisty is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.twisty. If not, see <http://www.gnu.org/licenses/>.

package smithers.twisty;

import smithers.java3d.Polygon3D;

/**
 * Represents a twisty puzzle such as a Rubik's Cube. In all of the methods in
 * this, a <emph>sticker index</emph> uniquely determines a sticker, not a
 * location. That is, after a twist is made, the {@code getColor(int)} method
 * should return the same values as before, but {@code getPolygon(int)} method
 * will not. If a puzzle has unstickered faces, it should declare those as
 * having stickers with a transparent colour.
 */
public interface TwistyPuzzle {
    
    /**
     * Gets the number of stickers on the puzzle.
     * 
     * @return the number of stickers
     */
    public int stickerCount();
    
    /**
     * Gets the colour of the specified sticker. The colour returned is
     * specified as ARGB.
     * 
     * @param index the sticker index
     * @return the colour of the sticker
     */
    public int getColor(int index);
    
    /**
     * Returns the absolute scale of the 3-d model of the puzzle. Essentially,
     * this should return the recommended 3-d width of the viewing window.
     * 
     * @return the scale
     */
    public double getScale();
    
    /**
     * Gets the 3-d polygon associated with the specified sticker.
     * 
     * @param index the sticker index
     * @return the polygon of the sticker
     */
    public Polygon3D getPolygon(int index);
    
    /**
     * Gets the twist instance determined by the sticker clicked on, and the
     * directon to twist. This should always return an instance which will be
     * accepted by {@link #twist(smithers.twisty.Twist)} (i.e. will not throw
     * {@code ClassCastException}, although the move may not be possible). The
     * {@code dir} argument should be: 0 for anticlockwise, 1 for clockwise or
     * 2 for a half twist.
     * 
     * @param index the sticker which was clicked on
     * @param dir   the directon
     * @return the twist instance
     */
    public Twist getTwist(int index, int dir);
    
    /**
     * Applies a twist.
     * 
     * @param twist the twist
     */
    public void twist(Twist twist);
    
    /**
     * Scrambles the puzzle.
     */
    public void scramble();
    
    /**
     * Works out whether this puzzle is in the solved state.
     *
     * @return {@code true} iff the puzzle is solved
     */
    public boolean isSolved();
    
}
